function [fmap info1] = getFieldmap(filename1, filename2, dononlinfmap);

%Generate fieldmap from gr images.
%
%Usage: calcFieldmap(image_type, filename1, filename2, delta_TE, imsize,  num_slice, num_coil);
%
%Input: filename1
%       filename2    
%
%Author: Youngkyoo Jung
%
%Date:   Feb. 19, 2008
%
%Changes: Valur Olafsson, Jan 20 2010.

PI = 3.142592;

%read information from DICOM header
fname1 = strcat(filename1,'/i00001.CFMRI.1');
info1 = dicominfo(fname1);  

if strcmp(filename1, filename2)
    fname2 = strcat(filename2,'/i00004.CFMRI.4');
    info2 = dicominfo(fname2);
    info2.EchoTime = info1.EchoTime + info2.Private_0019_10a8; % delta_TE is in 0019_10a8
    nimperslice = 6;
    fm1realsl_offset = 2;
    fm1imagsl_offset = 3;
    fm2realsl_offset = 5;
    fm2imagsl_offset = 6;
else
    fname2 = strcat(filename2,'/i00001.CFMRI.1');
    info2 = dicominfo(fname2);
    nimperslice = 4;
    fm1realsl_offset = 3;
    fm1imagsl_offset = 4;
    fm2realsl_offset = 3;
    fm2imagsl_offset = 4;
end

if (info1.Width ~= info2.Width) | (info1.Height ~= info2.Height) | ...
        (info1.ImagesInAcquisition ~= info2.ImagesInAcquisition)
   error('Image sizes of two acquisition are not matched!\n');
   exit;
end
freq_offset = 0;
if (info1.ImagingFrequency ~= info2.ImagingFrequency)
   fprintf('Center frequency of two acquisition are not matched!\n');
   freq_offset = (info2.ImagingFrequency - info1.ImagingFrequency)*1e6;
   fprintf('The freq. difference of %d Hz will be compensated!\n',int8(freq_offset));
end 


imsize_x =  info1.Width;
imsize_y = info1.Height;
num_slice = info1.ImagesInAcquisition ./ nimperslice;
delta_TE = (info2.EchoTime - info1.EchoTime);
      
Sum = zeros(imsize_x,imsize_y,num_slice);

%read images and composite 3D complex volumes from two images
for slice = 1:num_slice

    temp = fm1realsl_offset + (slice - 1)*nimperslice + 1000; temp_str = int2str(temp); slice_str = temp_str(2:4);
    fname1 = strcat(filename1,'/i00',slice_str,'.CFMRI.',int2str(fm1realsl_offset + (slice - 1)*nimperslice));
    temp_str = int2str(temp+1); slice_str = temp_str(2:4);
    fname2 = strcat(filename1,'/i00',slice_str,'.CFMRI.',int2str(fm1imagsl_offset + (slice - 1)*nimperslice));


    real1 = double(dicomread(fname1));
    imag1 = double(dicomread(fname2));

    image1(:,:,slice) = reshape(complex(real1, imag1), imsize_x, imsize_y);

    temp = fm2realsl_offset + (slice - 1)*nimperslice + 1000; temp_str = int2str(temp); slice_str = temp_str(2:4);
    fname1 = strcat(filename2,'/i00',slice_str,'.CFMRI.',int2str(fm2realsl_offset + (slice - 1)*nimperslice));
    temp_str = int2str(temp+1); slice_str = temp_str(2:4);
    fname2 = strcat(filename2,'/i00',slice_str,'.CFMRI.',int2str(fm2imagsl_offset + (slice - 1)*nimperslice));


    real2 = double(dicomread(fname1));
    imag2 = double(dicomread(fname2));

    image2(:,:,slice) = reshape(complex(real2, imag2), imsize_x, imsize_y);

    fclose('all');
end
Sum = Sum + image1.*conj(image2*exp(-i*freq_offset*2*pi*delta_TE*1e-3)); %compensate freq. mismatch.


Max = sqrt(max(max(max(abs(Sum)))));



im_phase = zeros(imsize_x,imsize_y,num_slice);
if dononlinfmap
    if 0  % estimate field map slice-by-slice
        fm_fmax = 0.1; fm_pl = 1; fm_init = [];
        fm_penord = [1 2 2]; fm_log2beta = [4 0 -2]; fm_niter = [50 100 200];
        fm_len = length(fm_penord);
        fm_reps = 3;  % # of field map repetitions

        for ii = 1:fm_reps
            for kk = 1:num_slice
                if sum(sum(abs(im_phase(:,:,kk)))) > 0
                    fm_init = im_phase(:,:,kk);
                end
                fm_indx = min(ii, fm_len);
                
                fmsm = mri_phase_denoise(Sum(:,:,kk), 'l2b', fm_log2beta(fm_indx), ...
                    'order', fm_penord(fm_indx), 'niter', fm_niter(fm_indx), ...
                    'wi_ml', true, 'pl', fm_pl, 'fmax', fm_fmax, 'init', fm_init);
                im_phase(:,:,kk) = fmsm;
            end
        end
    else  % estimate field map from 3D volume (default)
        fm_fmax = 0.1; fm_pl = 1; fm_init = [];
        fm_penord = [1 2 2]; fm_log2beta = [4 0 -2]; fm_niter = [50 50 300];
        fm_len = length(fm_penord);
        fm_reps = 3;  % # of field map repetitions

        for ii = 1:fm_reps
            if sum(sum(sum(abs(im_phase)))) > 0
                fm_init = im_phase;
            end
            fm_indx = min(ii, fm_len);
            
            fmsm = mri_phase_denoise(Sum, 'l2b', fm_log2beta(fm_indx), ...
                'order', fm_penord(fm_indx), 'niter', fm_niter(fm_indx), ...
                'wi_ml', true, 'pl', fm_pl, 'fmax', fm_fmax, 'init', fm_init);
            im_phase = fmsm;
        end
    end
else
    for ii = 1:imsize_x
        for jj = 1:imsize_y
            for kk = 1:num_slice
                if sqrt(abs(Sum(ii,jj,kk))) > Max/50
                      im_phase(ii,jj,kk) = angle(Sum(ii,jj,kk));    %angle(imag1) - angle(imag2);
                end
            end
        end
    end
end

fmap = im_phase/(2*pi*delta_TE)*1000;          %Convert radian to Hz
figure;
fmap = flipdim(permute(fmap, [2 1 3]),1);
smontage(fmap, 4,5,[-1/(2*delta_TE)*1000 1/(2*delta_TE)*1000]); title('Generated field map');colormap('jet');
fmap = fmap*2*pi;
%
return;
