function [H,D,U,OPT] = setup(kx,ky,nr)
%
% Set up interpolation parameters for gridding/SENSE.
% Trajectory must be scaled from 0 to +/- nr/2. Points
% outside this range are excluded from the reconstruction.
%
% Output:
%  H is the interpolation matrix
%  D is a diagonal density weighting matrix
%  U is the deapodization function
%  OPT is a structure of SENSE options
%
% Inputs:
%  kx,ky are k-space indices
%  nr is the resolution (matrix dimension)
%
%(c) Mark Bydder (submitted to Mag Res Imaging 2006)

% SENSE options
WIDTH = 4;								% kernel width
OSF = 2.0;								% oversampling factor
X=[1 1.5 2 3];Y=[1.5 2.05 2.34 2.6];	% Kaiser Bessel parameter
SIGMA = WIDTH*spline(X,Y,OSF);			% (Fessler et al, IEEE Signal Proc 2003;51:560)
ITERS = 10;								% iterations (0=conventional sum of squares)
DAMP = 0.01;                            % damping factor (regularization)
LAMB = 0.0;								% phase constraint (partial Fourier)
KERNEL = 16;							% resolution of coil sensitivities

% create structure
OPT = struct('width',WIDTH,'osf',OSF,'sigma',SIGMA,'iters',ITERS,'damp',DAMP,'lamb',LAMB,'kernel',KERNEL);

% error check
m = numel(kx);
if numel(ky)~=m
    error('kx,ky size mis-match')
end

% size of scaled matrix (must be even)
n = 2*ceil(OPT.osf*nr/2);

% exclude out of bound and scale kspace
nz = find(abs(kx)<nr/2 & abs(ky)<nr/2);
kx = kx(nz)*n/nr;
ky = ky(nz)*n/nr;
p = length(nz);
if p~=m
	disp(['WARNING! ' num2str(m-p) ' points (out of ' num2str(m) ') are out-of-bounds.'])
end

% radius of convolution kernel
radius = ceil((OPT.width-1)/2);

% indices for sparse matrix
ya = zeros(p,(2*radius+1)^2);
xa = zeros(p,(2*radius+1)^2);
sa = zeros(p,(2*radius+1)^2);

for j = -radius:radius
  for k = -radius:radius
		% indices
	  	x = round(kx+j);
	  	y = round(ky+k);
		index = (j+radius)*(2*radius+1)+k+radius+1;
		% convolution weighting
        sa(:,index) = kaiser_bessel_zero(kx-x,OPT.width,OPT.sigma)...
                    .*kaiser_bessel_zero(ky-y,OPT.width,OPT.sigma);
		% wrap around
		x = mod(x,n);
		y = mod(y,n);
		% store indices (NB. MATLAB unit offset)
		xa(:,index) = 1+x+n*y;
		ya(:,index) = nz;
	end
end

% normalization for area under convolution kernel
c = kaiser_bessel_zero(-radius:radius,OPT.width,OPT.sigma);
scale = 1/sum(kron(c,c));

% make matrix structure
H = sparse(ya,xa,scale*sa,m,n^2);

% data weighting preconditioner
if nargout>1
    d = -1;
    w = scale;
    while any(d<0)
        d = density(H,w,10);
        w = w+scale;
    end
    D = sparse(1:m,1:m,d);
end

% deapodization function
if nargout>2
    U = 1./abs(fft([realsqrt(scale)*c zeros(1,n-length(c))]));
    U = U'*U;
end
