 function [corr_image, td] = spiralfmap2(Pfile_name, fmapdir1, fmapdir2, output, skip_slice, ASL, dotoeplitz, parallel)
%function [corr_image, td] = spiralfmap2(Pfile_name, fmapdir1, fmapdir2, output, skip_slice, ASL, dotoeplitz, parallel)
%
%      Off-resonance correction program for spiral acquisition
%      using field map
%        
%      Ver: 1.3
%
  %usage: corr_image = spiralfmap(Pfile_name, fmap1, fmap2, outputname, skip_slice, ASL)
%
% Input: 
%       Pfile_name: P-file name of images to be corrected
%       fmapdir1:   the directory name at TE1
%       fmapdir2:   the directory name at TE2
%       output: AFNI-BRIK name for corrected image (optional)
%                   default name is Pfile_name+dewarped.      
%       skip_slice: number of slices to be skipped from fieldmap
%                   (optional)          
%       ASL:        option for ASL or fmri (signal averaging for ASL) (optional)
%                   default is 0 (no signal averaging).
%       dotoeplitz: [use Toeplitz recon, target fwhm]   (default: [true, 1.4])
%       parallel:   use parallel code (currently not implemented)
%
%
% Output:
%       field map corrected image
%
%  Author: Youngkyoo Jung, PhD
%          fMRI center, Radiology, UC San Diego
%
%  This program is based on the algorithm from the article:
%     1) B.P. Sutton and J.A. Fessler and D.C. Noll.  Fast, iterative, 
%            field-corrected image reconstruction for MRI.   
%            IEEE Trans. Med. Im., 2003, 22(2), p 178-188
%     2) J.A Fessler and Sangwoo Lee and V.T. Olafsson and H.R. Shi
%            and Noll, D.C.  Toeplitz-Based Iterative Image
%            Reconstruction for MRI With Correction for Magnetic 
%            Field Inhomogeneity.
%            IEEE Trans. Sig. Proc., 2005, 53(9), p.3393-3402
%
% Version      date       author      description
%    1.0    8/25/2008       YJ       Implementation of field map
%                                    correction program for spiral
%                                    acquisition
%    1.1    9/03/2008       YJ       Added codes for dual-echo acq.
%    1.2    10/29/2008      YJ       Added codes for multiple interleaves.
%    1.3    10/12/2009      YJ       Added codes for negative skip_slice
%                                    (Image has more slices than fmap).
%    1.4    01/11/2010      VO       Added faster Toeplitz based code path
%
%
fullloc = which(mfilename);
loc=fullloc(1:(end-(length(mfilename)+2)));

% add paths
currpath = path;  % save current path to restore later
addpath(sprintf('%sirt/general', loc), '-begin');
addpath(sprintf('%sirt/mri', loc), '-begin');
addpath(sprintf('%sirt/systems', loc), '-begin');
addpath(sprintf('%sirt/wls', loc), '-begin');
addpath(sprintf('%sirt/penalty', loc), '-begin');
addpath(sprintf('%sirt/nufft', loc), '-begin');
addpath(sprintf('%sirt/mex/v7', loc), '-begin');
addpath(sprintf('%sirt/utilities', loc), '-begin');


%--------------------------------------
% check and convert input parameters
%--------------------------------------


if ~exist('skip_slice','var') || isempty(skip_slice)
   skip_slice = 0;
end

if ~exist('ASL','var') || isempty(ASL)
   ASL = 0;
end

if ~exist('dotoeplitz','var') || isempty(dotoeplitz)
    dotoeplitz = true;
    tfwhm = 1.4;
elseif length(dotoeplitz) < 2
    tfwhm = 1.4;
elseif length(dotoeplitz) == 2
    tfwhm = dotoeplitz{2};
    dotoeplitz = dotoeplitz{1};
else
    error 'dotoeplitz argument is wrong'
end

if ~exist('parallel','var') || isempty(parallel)
    parallel = false;
end

%Generate field maps
[fmap dicom_info] = getFieldmap(fmapdir1, fmapdir2, dotoeplitz);

% check if 20x
if length(findstr(dicom_info.SoftwareVersion, '20')) > 0
    is20x = true;
else
    is20x = false;
end

%Get scan parameters from AFNI or generate AFNI
h = getgrheader(Pfile_name, is20x);

if h.necho == 1
    if ~exist('output') || isempty(output)
        output = Pfile_name;
        brikname = [Pfile_name,'brik+orig'];   
        outputname = strcat(Pfile_name,'brik+dbl');
    else
        brikname = [output,'brik+orig']; 
        outputname = strcat(output,'brik+dbl');
    end
else
    if ~exist('output') || isempty(output)
        output = Pfile_name;
        brikname = [Pfile_name,'brik_e01+orig'];
        outputname(:,1) = strcat(Pfile_name,'brik_e01+dbl');
        outputname(:,2) = strcat(Pfile_name,'brik_e02+dbl');
    else
        brikname = [output,'brik_e01+orig'];
        outputname(:,1) = strcat(output,'brik_e01+dbl');
        outputname(:,2) = strcat(output,'brik_e02+dbl');
    end
end


if ~system(['ls ', brikname, '*'])
    h = parseAFNIhead(h, brikname, {'FOV', 'nPix','TE','mpsFREQ'});
else     %No brik file, generate afni file
    rungr(Pfile_name, output, is20x);
    h = parseAFNIhead(h, brikname, {'FOV', 'nPix','TE','mpsFREQ'});
end    

if dicom_info.ImagingFrequency*1e7 ~= h.mpsFREQ
    fprintf('Center frequency of fieldmap and acquisition are not matched!\n');
    fprintf('cf of field map is %fHz, cf of acquisition is %fHz!\n', ... 
                dicom_info.ImagingFrequency*1e6,h.mpsFREQ/10);
    cf_diff = dicom_info.ImagingFrequency*1e6 - h.mpsFREQ/10; %unit of Hz
    fmap = fmap + cf_diff*2*pi;
end

%Setup scan parameters
FOV = h.FOV;   %in cm
N = h.nPix;     %reconstruction size
TE = h.TE*1e-03;  %TE in seconds
nsamp = h.nsamp;  %number of samples per interleave
nl = h.nl;   %number of interleaves
nreps = h.reps;
nslice = h.nslices;
ncoil = h.ncoil;
necho = h.necho;
    
%In case of negative skip_slice
if skip_slice < 0
    temp_fmap = zeros(N,N,nslice);
    temp_fmap(:,:,abs(skip_slice)+1:abs(skip_slice)+size(fmap,3)) = fmap;
    clear fmap;
    fmap = temp_fmap;
    clear temp_fmap;
    skip_slice = 0;
end

%match image location to gr (unit of pixels)
h.xoff = h.xoff - 1/2;
h.yoff = h.yoff + 1/2;
    
i = sqrt(-1);

rBW = 125; %all spiral scan has rBW of 125 kHz

[kx ky] = getkspacedata(Pfile_name, nl, is20x);

[H,D,U,OPT] = setup(kx,ky,N);

ww = diag(D);

clear H, clear D, clear U, clear OPT;

niter = 6;

ndat = length(ww)/nl;    %num data pts for one spiral

% Set up timing of slice acquisition
sampling_rate = 1/(2*rBW*1000);

tt = zeros(ndat,necho);
tt_ext = zeros(ndat*nl,necho);
for ii = 1:necho
    tt(:,ii) = [(0:(ndat-1))*sampling_rate+TE(ii)]'; %Timing of non-delayed acquisition

    for mm = 1:nl %for multi-shot case
         tt_ext(1+(mm-1)*size(tt,1):size(tt,1)*mm,ii) = [tt(:,ii)];
    end

end

    
if dotoeplitz
    % common setup options for recon objects
    kspace = [ky kx] ./ N;  % make certain FOV = [N N] for Gmri
                            % and need to switch kx ky to make image space coordinates consistant
    nufft_args = {[N N], [6 6], 2*[N N], ...
                  [N N]./2 - 1/2, 'table', 2^12, 'minmax:kb'};
    basis = {'rect'};
    Lfind = {1,5e-3};
    maxniter = niter;
    stop_thr = 1e-3;

    % Set up mask
    [xx, yy] = ndgrid([-N/2:N/2-1]+1/2);
    mask = logical(abs(xx + 1i*yy) < N/2-1);

    % make basic fast object and penalty
    A_fastbase = Gmri(kspace, mask, 'fov', [N N], 'basis', basis, ...
                  'nufft', nufft_args);
    R = geniterrecpen(A_fastbase, tfwhm, dotoeplitz);
else
    %Set up mask
    mask = logical(ones(N,N));

    % PENALTY weighting for penalized CGM
    Cbeta = 2^9;        % seems a bit high??
    R = Robject(mask, 'edge_type', 'leak', 'order', 1, 'beta', Cbeta);
    % % PENALTY weighting for penalized CGM
    % C = C2sparse('leak', mask, 8, 0);
    % C = sqrt(2^(9)) * C;   % 2^(10)
end


start_time = clock;

if ASL == 1
    raw = zeros(h.nsamp,nl,ncoil,2,necho);
else
    raw = zeros(h.nsamp,nl,ncoil,nreps,necho);
end

for slice = 1:nslice
    
    fprintf('Slice: %d\n', slice);
    raw_temp=getrawdata(Pfile_name,slice,nreps,is20x);
    if ASL == 1
      raw(:,:,:,1,:) = mean(raw_temp(:,:,:,1:2:nreps,:),4);
      raw(:,:,:,2,:) = mean(raw_temp(:,:,:,2:2:nreps,:),4);
    else
        raw = raw_temp;
    end
    
    if ~dotoeplitz
        raw = fermi_filter(kx,ky,raw,h);   %fermi filtering on the k-space edge
    end

    if slice == 1
       corr_image = zeros(N,N,nslice,size(raw,4),necho); 
       temp_image = zeros(N,N,ncoil,size(raw,4),necho);
    end
    
    %put linear phase offset on raw data for off-axis imaging
    phase_off = (h.xoff*kx + h.yoff*ky)/(h.nPix)*2*pi;
    phase_off = reshape(phase_off, nsamp,nl);
    raw = bsxfun(@times, raw, exp(1i*phase_off));
    clear phase_off;
        
    fmap_slice = fmap(:,:,slice + skip_slice);
    for echo = 1:necho
        if dotoeplitz
            zmap  = 1i*fmap_slice;
            A_fast = A_fastbase.arg.new_zmap(A_fastbase, tt_ext(:,echo), ...
                          zmap, Lfind);
            % A_fast = Gmri(kspace, mask, 'fov', [N N], 'basis', basis, ...
            %               'nufft', nufft_args, 'ti', tt_ext(:,echo), ...
            %               'zmap', zmap, 'L', Lfind);
            
            if A_fast.arg.aL > 9
                % set a tolerance when calculating exp() approximations
                A_fast.arg.exp_approx_args = {'tol', {'fro'}};

                % recalculate using new tolerance
                A_fast = A_fast.arg.new_zmap(A_fast, tt_ext(:,echo), zmap, ...
                        A_fast.arg.L, A_fast.arg.aL);
            end

            A_fastn = A_fast.arg.new_image_basis(A_fast, {'dirac'});
            T_fast = build_gram(A_fast, [], []);  % Toeplitz object
        else
            A_fast = fast_mr(kx,ky,FOV,N,2*N,4,tt_ext(:,echo),fmap_slice(:),1,10*nl,1);
            A_fastn = fast_mr(A_fast,'flgswt',0);
        end
        
        for coil = 1:ncoil
            tstart = cputime;
            for reps = 1:size(raw,4)
                if mod(reps,20) == 0
                    fprintf('.');
                end
                kdata = reshape(raw(:,:,coil,reps,echo), nl*nsamp, 1);

                it_init = (A_fastn' * (full(ww).*kdata)) ./ (N.^2);
                if dotoeplitz
                    bb = A_fast' * kdata;
                    it_fast = qpwls_pcg2(it_init, T_fast, bb, R.C, ...
                            'niter', maxniter, 'isave', 'all', ...
                            'stop_threshold', stop_thr);
                    it_fast = embed(it_fast, mask(:));
                else
                    it_fast = qpwls_pcg(it_init, A_fast, 1, kdata, 0, R.C, 1, niter+1, mask, 0);
                end

                temp_image(:,:,coil,reps,echo) = reshape(it_fast(:,end),64,64);
            end %repetition loop
            td(coil,echo,slice) = cputime - tstart;
        end  %coil loop
    end  %echo loop
    fprintf('\n');
    corr_image(:,:,slice,:,:) = sqrt(sum(abs(temp_image).^2,3));    %sum of square method

end  %slice loop

end_time = clock;
elapsed = etime(end_time, start_time);

clear temp_image, clear orig;

fprintf('Total Elapsed time is %dhours %dminutes %dseconds\n',floor(floor(elapsed/60)/60),mod(floor(elapsed/60),60),int8(mod(elapsed,60)));    

if necho == 1
    figure;
    title('Field map corrected Images')
    smontage(mean(corr_image,4), 4,5);
else
    for ii = 1:necho
        figure;
        title(sprintf('Field map corrected Images from echo %d',ii));
        smontage(mean(corr_image(:,:,:,:,ii),4), 4,5);
    end
end
corr_image = flipdim(corr_image,1);

%make circular FOV
for ii = 1:size(corr_image,1)
    for jj = 1:size(corr_image,2)
        rad = sqrt((ii - h.nPix/2)^2 + (jj - h.nPix/2)^2);
        if rad > size(corr_image,1)/2
            corr_image(ii,jj,:,:,:) = 0;
        end
    end
end

%Generating AFNI-BRIK file
if necho == 1
    mat2brik([],corr_image(:,:,:,:,1),outputname,brikname(1:end-5),256);
else
    for ii = 1:necho
        mat2brik([],corr_image(:,:,:,:,ii),outputname(:,ii)',strcat(brikname(1:end-8),sprintf('e0%d',ii)),256);
    end
end

if necho == 1
    command = ['cp ',brikname,'.HEAD ',outputname,'+orig.HEAD'];
    system(command);
    if ASL == 1
        command1 = sprintf('3drefit -atrint TAXIS_NUMS ''2 %d 77002'' %s+orig', size(corr_image,3), outputname);
        command2 = sprintf('3drefit -atrint DATASET_RANK ''3 2'' %s+orig', outputname);
        system(command1);
        system(command2);
    end
else
    for ii = 1:necho
        command = ['cp ',strcat(brikname(1:end-8),sprintf('e0%d',ii),'+orig'),'.HEAD ',outputname(:,ii)','+orig.HEAD'];
        system(command);   
        if ASL == 1
            command1 = sprintf('3drefit -atrint TAXIS_NUMS ''2 %d 77002'' %s+orig', size(corr_image,3), outputname(:,ii));
            command2 = sprintf('3drefit -atrint DATASET_RANK ''3 2'' %s+orig', outputname(:,ii));
            system(command1);
            system(command2);
        end
    end
end

% reset path
path(currpath);
